/*************************************************************
*	PROGETTO LABIRINTO
*
* nome file: <labirinto.c>
*
* Autori:
* Thomas Carrabs
* Roberto Pisoni
*
* Progetto realizzato per l'esame di Geometria Computazionale
* e di Informatica Grafica presso il Dipartimento di Scienze
* dell'Informazione dell'universit Statale di Milano
*
* Tutto il codice  distribuito sotto licenza Gnu/GPL
* http://www.gnu.org/copyleft/gpl.html
* una copia della licenza si trova nella cartella License del
* Progetto
*
*************************************************************/


/*
 * In questo file viene creata la struttura matematica del labirinto,
 * la sua soluzione.
 * Viene poi chiamata la funzione geometria() che si occupa di visualizzare tutto questo.
 */
#include "labirinto.h"
#include "stack.h"
#include "coda.h"
#include "geometria.h"

struct maze maze;
stack stk;
int *soluz_lab;

int main(int argc,char *argv[])
{
/*
 * Crea la struttura matematica del labirinto (grafo) in base al numero di righe e di colonne
 * passate al momento di lancio del programma.
 */
	int dimens,errore;

	switch(argc) //controllo sul numero di parametri di input
	{
		case 1:
			fprintf(stderr, "Errore! Mancanza di parametri.\nUSO: %s <linee> <colonne>\n", argv[0]);
			exit(1);
		case 2:	
		  fprintf(stderr, "Errore! Mancanza parametro numero di colonne.\nUSO: %s <linee> <colonne>\n", argv[0]);
			exit(1);
	}	
	maze.m=atoi(argv[1]); //conversione parametri di input da ascii a int
	maze.n=atoi(argv[2]);
	maze.l=10;			  //settiamo di default la dimensione della cella a 10 pixel.
	errore=0;
	if(maze.m <=1)			//controllo sul valore dei parametri di input
		errore=errore+1;
	if(maze.n <=1)
		 errore=errore+2;
	switch(errore)
	{
		case 1:
			fprintf(stderr,"Errore! Inserire un numero di righe maggiore o uguale a 2\n");
			exit(1);
		case 2:
 			fprintf(stderr,"Errore! Inserire un numero di colonne maggiore o uguale a 2\n");
			exit(1);
		case 3:
			fprintf(stderr,"Errore! Inserire un numero di righe e colonne maggiore o uguale a 2\n");
			exit(1);
	}
	
	srand(time(NULL));
	dimens=maze.m*maze.n;	//dimensione del labirinto: nro righe*nro colonne
	maze.pcella = calloc(dimens,sizeof(struct cella));	//inizializza la lista di adiacenza
	maze.vicini = calloc(dimens,sizeof(int));		//vettore contatore vicini lista d'adiacenza

	if(!crealista(maze.pcella,dimens,maze.m,maze.n, maze.vicini)){
		fprintf(stderr,"Errore nella creazione della lista concatenata.\n");
		exit(1);
	}

	if(!creaLabirinto(&maze,dimens)){
		fprintf(stderr,"Errore nella creazione del labirinto.\n");
		exit(1);
	}
	if((soluz_lab=soluzione(&maze))==NULL){
		fprintf(stderr,"Errore nel determinare la soluzione del labirinto.\n");
		exit(1);
	}

	geometria(argc, argv);

	return 0;
}

short int crealista(cella *ptrcella,int dimensione,int rig,int col, int *vic)
{
/*
 * Crea la lista di adiacenza per le celle del labirinto
 */
	
	int i,x,y;
	cella *temp,*curr;
	
	for(i=0;i<dimensione;i++){
		ptrcella[i].id = i;		//aggiorno id cella nella lista di adiacenza
		x=i%col;				//calcolo la x corrispndente all'indice i
		y=i/col;				//calcolo la y corrispndente all'indice i
		curr=&ptrcella[i];
		if(x-1>=0) /*esiste la cella a sinistra?*/
		{
			if((temp=calloc(1,sizeof(cella)))==NULL)
			   return 0; //c' stato un errore
			temp->id=y*col+x-1;		//calcolo id della cella a sinistra
			temp->next_cella=NULL;	// l'ultimo elemento inserito quindi non c' elemento successivo
			curr->next_cella=temp;	//la cella precedente punta a quella appena creata
			curr=temp;				//la cella attuale diventa quella appena creata
			vic[i]++;				//la cella di indice i ha un vicino in pi
		}
		if(y+1<rig) /*esiste la cella sopra?*/
		{
			if((temp=calloc(1,sizeof(cella)))==NULL)
			   return 0; //c' stato un errore
			temp->id=(y+1)*col+x;
			temp->next_cella=NULL;
			curr->next_cella=temp;
			curr=temp;
			vic[i]++;
		}
		if(x+1<col) /*esiste la cella a destra?*/
		{
			if((temp=calloc(1,sizeof(cella)))==NULL)
			   return 0; //c' stato un errore
			temp->id=y*col+x+1;
			temp->next_cella=NULL;
			curr->next_cella=temp;
			curr=temp;
			vic[i]++;
		}		
		if(y-1>=0) /*esiste la cella sotto?*/
		{
			if((temp=calloc(1,sizeof(cella)))==NULL)
			   return 0; //c' stato un errore
			temp->id=(y-1)*col+x;
			temp->next_cella=NULL;
			curr->next_cella=temp;
			curr=temp;
			vic[i]++;
		}
	}
	return 1;
}

int cellaBordo(int m,int n){
/*
 * Scegli una cella su uno dei 4 bordi del labirinto.
 * questa cella sar la cella di ingresso nel labirinto
 */
	int x,y,temp;
	
	temp=rand();
	x=temp%n; //scelta colonna a caso
	if((x==0)||(x==n-1))
	{
		y=temp%m; //scelta riga a caso
	}
	else{
		if((temp%2)==0) //scelta 1a riga o ultima riga controllando se il numero
						//random e' pari o dispari
			y=0;
		else
			y=m-1;
	}
	return (y*n+x);
//	return 0; //cella utilizzata per il debug ;-)
}

short int creaLabirinto(struct maze *lab, int dimensione){
/*
 * Crea il labirinto
 */
	int curr,celle_tot,nro_vicini,scelta,i;
	int x,y,x_1,y_1;
	struct cella_vis *vis;
	cella *temp;
	
	//vettore che mi indica se i-esima cella  stata visitata
	//0 no 1 si
	vis=calloc(dimensione,sizeof(struct cella_vis));
	
	creaStack(&stk);
	

	curr=cellaBordo(lab->m,lab->n);
	lab->entrata = &(lab->pcella[curr]);

	vis[curr].vis = 1;
	

	celle_tot=1;

	//AGGIORNARE I CONTATORI DELLA PRIMA CELLA;
	aggiornaVicini(lab, curr, vis);
	pushStack(&stk,lab->entrata);

	while(celle_tot<dimensione)
	{
		nro_vicini=lab->vicini[curr];
	
		temp=lab->pcella[curr].next_cella; //mi sposto sulla prima cella vicina: in questo caso scelta  uguale a 0
		scelta=rand()%nro_vicini; //scelgo a caso una cella tra quelle vicine
	
		// con questo ciclo for mi sposto sulla cella vicina a random
		// utilizzando l'aritmetica dei puntatori...
		for(i=1;i<=scelta;i++)
		{
			temp=temp->next_cella;
		}
							
		if(vis[temp->id].vis== 0)	//la cella CANDIDATA non  stata ancora visitata
		{
			if(vis[lab->pcella[curr].id].nvic >= nro_vicini){ //se il numero dei vicini occupati della cella chiamante la candidata  uguale al numero massimo dei vicini eseguo un pop
				curr = popStack(&stk);
			}
			else{
			
				pushStack(&stk,temp); //sullo stack cella corrente
			
				x=curr%lab->n;				   //creo corridoio
				y=curr/lab->n;
				x_1=temp->id%lab->n;
				y_1=temp->id/lab->n;
				/*NORD +1 OVEST +2 SUD +4 EST +8*/
				if(x<x_1)	//creo corridoio verso EST nella cella x verso OVEST cella x_1
				{
					lab->pcella[curr].passaggio=lab->pcella[curr].passaggio+8;
					lab->pcella[temp->id].passaggio=lab->pcella[temp->id].passaggio+2;
				}
				if(x>x_1)
				{   //creo corridoio verso OVEST nella cella x verso EST cella x_1
					lab->pcella[curr].passaggio=lab->pcella[curr].passaggio+2;
					lab->pcella[temp->id].passaggio=lab->pcella[temp->id].passaggio+8;
				}
				if(y<y_1) //creo corridoio verso NORD nella cella x verso SUD nella cella x_1
				{
					lab->pcella[curr].passaggio=lab->pcella[curr].passaggio+1;
					lab->pcella[temp->id].passaggio=lab->pcella[temp->id].passaggio+4;
				}
				if(y>y_1) //creo corridoio verso SUD nella cella x verso NORD nella cella x_1
				{
					lab->pcella[curr].passaggio=lab->pcella[curr].passaggio+4;
					lab->pcella[temp->id].passaggio=lab->pcella[temp->id].passaggio+1;
				}
				
				curr=temp->id;					   //aggiorno cella corrente
				vis[curr].vis = 1;				   //marcata come visitata
				celle_tot++;

				//AGGIORNO CONTATORI CELLE VICINE
				aggiornaVicini(lab,curr,vis);
			}
		}
		else
		{

			if(vis[curr].nvic == lab->vicini[curr]){
				curr = popStack(&stk);
			}
		}

	}

	lab->uscita = &(lab->pcella[cellaUscita(lab->entrata->id)]);
	//printf("ID_uscita: %d\n", lab->uscita->id);
	return 1;
}


void aggiornaVicini(struct maze *lab, int ID, struct cella_vis *vis){
//AGGIORNO CONTATORI VICINI CELLA

	struct cella *ptmp;
	int i,id_tmp;

	ptmp = &lab->pcella[ID];
	for(i=0; i<lab->vicini[ID];i++){
		ptmp = ptmp->next_cella;
		id_tmp = ptmp->id;
		if(id_tmp != ID){
			vis[id_tmp].nvic++;
		}
	}
}

int cellaUscita(int ID){
/*
 * Scegli una cella di uscita sul bordo del opposto a quello di entrata
 */
	int xp, yp, xt, yt;

	xp = ID%maze.n; //scelta colonna
	yp = ID/maze.n; //scela riga 


	srand(time(NULL));

	if(xp == 0){
		xt = maze.n-1;
	}
	else{
		if(xp == maze.n-1){
			xt = 0;
		}
		else{
			//scelgo a caso xt
			xt = (int) ((maze.n-1)*(rand()/(RAND_MAX+1.0)));
		}
	}
	
	if(yp == 0){
		yt = maze.m-1;
	}
	else{
		if(yp == maze.m-1){
			yt = 0;
		}
		else{
			//scelgo a caso yt
			yt = (int) ((maze.m-1)*(rand()/(RAND_MAX+1.0)));
		}
	}

	return (yt*maze.n+xt);
}

int *soluzione(struct maze *lab){
/*
 * Calcola la soluzione del labirinto
 * Il percorso verso l'uscita  il cammino minimo dalla cella di ingresso a quella di uscita
 */
	int dim,i,temp,flag;
	typedef enum{bianco,grigio,nero} colore;
	int *pred;		//contiene il predecessore di ogni nodo
	colore *col;	//se ho visitato tutti i vicini di una cella	
	coda cod;
	cella cella_next;
	int x_next,y_next,xt,yt;

	dim=lab->m*lab->n;
	if((pred=calloc(dim,sizeof(int)))==NULL)
			   return 0; //c' stato un errore
	if((col=calloc(dim,sizeof(colore)))==NULL)
			   return 0; //c' stato un errore
	creaCoda(&cod);
	for(i=0;i<dim;i++){
		pred[i]=-1;	//inizializza la distanza e il predecessore di tutte le celle a -1
		col[i]=bianco;			//il colore a bianco (nessuna cella adiacente  stata visitata)
	}
	col[lab->entrata->id]=grigio;	//le celle adiacenti della cella d'entrata sono state visitate in parte
	mettiCoda(&cod,lab->entrata);	//pone cella d'entrata nella coda
	while(cod.ctr>0){					//finch ci sono elementi
		temp=cod.top->id;				//prendi elemento dalla coda (il 1o inserito) senza toglierlo
		cella_next=lab->pcella[temp];
		while(cella_next.next_cella!=NULL){   //per ogni cella adiacente a quella attuale aggiorno se....
			cella_next=*cella_next.next_cella;
			flag=0;
			//controllo che non ci sia un muro tra le celle
			x_next=cella_next.id%lab->n;
			y_next=cella_next.id/lab->n;
			xt=temp%lab->n;
			yt=temp/lab->n;
			if(x_next<xt){	//se cella  a sinistra di quella corrente deve esistere un passaggio aperto a destra
				if(lab->pcella[cella_next.id].passaggio<8)
					flag=1;	//se diverso passa ad un'altra cella
			}
			if(x_next>xt){	//se cella  a destra di quella corrente deve esistere un passaggio aperto a sinistra 
				if(lab->pcella[cella_next.id].passaggio!=2 && lab->pcella[cella_next.id].passaggio!=3 &&
					lab->pcella[cella_next.id].passaggio!=6 && lab->pcella[cella_next.id].passaggio!=7 &&
					lab->pcella[cella_next.id].passaggio!=10 && lab->pcella[cella_next.id].passaggio!=11 &&
					lab->pcella[cella_next.id].passaggio!=14)
					flag=1;	
			}
			if(y_next<yt){	//se cella  sotto quella corrente deve esistere un passaggio aperto a nord
				if(lab->pcella[cella_next.id].passaggio!=1 && lab->pcella[cella_next.id].passaggio!=3 &&
					lab->pcella[cella_next.id].passaggio!=5 && lab->pcella[cella_next.id].passaggio!=7 &&
					lab->pcella[cella_next.id].passaggio!=9 && lab->pcella[cella_next.id].passaggio!=11 &&
					lab->pcella[cella_next.id].passaggio!=13)
					flag=1;	
			}
			if(y_next>yt){	//se cella  sopra quella corrente deve esistere un passaggio aperto a sud
				if(lab->pcella[cella_next.id].passaggio!=4 && lab->pcella[cella_next.id].passaggio!=5 &&
					lab->pcella[cella_next.id].passaggio!=6 && lab->pcella[cella_next.id].passaggio!=7 &&
					lab->pcella[cella_next.id].passaggio!=12 && lab->pcella[cella_next.id].passaggio!=13 &&
					lab->pcella[cella_next.id].passaggio!=14)
					flag=1;	
			}
			if((flag==0) && (col[cella_next.id] == bianco)){	//la cella ha nessun vicino ancora visitato
				col[cella_next.id]=grigio;		//la cella ha dei vicini visitati
				pred[cella_next.id]=temp;		//aggiorno il predecessore
				mettiCoda(&cod,&cella_next);	//pongo nella coda la cella attuale	
			}
		}
		temp=togliCoda(&cod);	//tolgo dalla coda il 1o elemento inserito
		col[temp]=nero;		//della cella temp ho visitato tutte quelle adiacenti
	}
	free(col);
	return pred;
}

